#TODO change devies.yaml from arg to flag with default value

from __future__ import print_function

import argparse
import json
import os
import re
import sys
import yaml

from colorama import init as init_colorama

from consts import VERSION
from utils import log, info, warn, error, deep_merge
from yaml_utils import init_yaml, ordered_load, ordered_dump

class EspHomePreprocessorError(Exception):
  pass

class EspHomePreprocessorTemplateError(EspHomePreprocessorError):
  def __init__(self, file, line, message):
    super(EspHomePreprocessorError, self).__init__(message)

    self.file = file
    self.line = line

  def __str__(self):
    return self.file + ":" + str(self.line+1) + " " + self.message

def main():
  init_colorama()

  try:
    return run(sys.argv)
  except EspHomePreprocessorError as e:
    error(e)
    return 1
  except KeyboardInterrupt:
    return 1

def run(argv):
  args = parse_args(argv)

  if args.command == "version":
    return run_version(args)
  elif args.command == "build":
    return run_build(args)
  else:
    raise EspHomePreprocessorError("Unknown command: {}".format(args.command))

def parse_args(argv):
    parser = argparse.ArgumentParser(prog="ehdm")

    subparsers = parser.add_subparsers(help="Commands", dest="command")
    subparsers.required = True

    parser_version = subparsers.add_parser("version", help="Get version number.")

    parser_build = subparsers.add_parser("build", help="Build output.")

    parser_build.add_argument("-c", "--catalog", default="devices.yaml", help="Your YAML device catalog file.")
    parser_build.add_argument("device", help="The device to build from the catalog.")

    return parser.parse_args(argv[1:])

def run_version(args):
  print("EspHomePreprocessor version {}".format(VERSION))
  return 0

def run_build(args):
  init_yaml()

  log("Building device '{0}' from catalog '{1}'...".format(args.device, args.catalog))

  catalog = load_catalog(args.catalog)
  devices = load_devices(catalog)
  device = compute_device(catalog, devices, args.device)

  validate_device(device)

  template = load_template(device)

  output = process_template(template, device)

  save_output(output, device)

  return 0

def load_catalog(file):
  try:
    with open(file, "r") as stream:
      try:
        return ordered_load(stream, yaml.SafeLoader)
      except yaml.YAMLError as ex:
        raise EspHomePreprocessorError("Error parsing device catalog file: {0}".format(ex))
  except IOError as ex:
    raise EspHomePreprocessorError("Error loading device catalog file: {0}".format(ex))

def load_devices(catalog):
  try:
    return catalog["devices"]
  except KeyError as ex:
    raise EspHomePreprocessorError("Error computing device: Catalog has no 'devices' key")

def find_device(devices, device_id):
  try:
    device = devices[device_id]

    if device is None:
      raise EspHomePreprocessorError("Error computing device: Catalog has empty device '{0}'".format(device_id))

    return device
  except KeyError as ex:
    raise EspHomePreprocessorError("Error computing device: Catalog has no device '{0}'".format(device_id))

def compute_device(catalog, devices, device_id):
  hierarchy = []

  device = find_device(devices, device_id)
  hierarchy.append(device)

  while True:
    try:
      parent_device_id = device["inherits"]
      del device["inherits"]

      device = find_device(devices, parent_device_id)
      hierarchy.append(device)
    except KeyError as ex:
      break

  try:
    device = catalog["global"]
    hierarchy.append(device)
  except KeyError as ex:
    pass

  hierarchy.reverse()

  computed_device = dict()

  for device in hierarchy:
    try:
      undefines = device["undefines"]

      del device["undefines"]

      for undefine in undefines:
        try:
          del computed_device["defines"][undefine]
        except KeyError as ex:
          pass
    except KeyError as ex:
      pass

    deep_merge(computed_device, device)

  #Copy device ID to defines
  computed_device["defines"]["id"] = device_id

  return computed_device

def validate_device(device):
  try:
    device["template"]
  except KeyError:
    raise EspHomePreprocessorError("Device has no template")

  try:
    device["output_directory"]
  except KeyError:
    raise EspHomePreprocessorError("Device has no output directory")

def load_template(device):
  f = open(device["template"], "r")
  template = f.readlines()
  f.close()

  return template

def process_template(template, device):
  output = []
  ifblocks = []
  keep = True

  for linenumber, line in enumerate(template):
    str = line.strip()

    if str.startswith("#rem"):
      pass
    elif str.startswith("#ifdef") or str.startswith("#ifndef"):
      parts = str.split()

      if parts[0] == "#ifdef" and (parts[1] in device["defines"]):
        keep = True
      elif parts[0] == "#ifndef" and (parts[1] not in device["defines"]):
        keep = True
      else:
        keep = False

      ifblocks.append({"text": str, "linenumber": linenumber+1, "directive": parts[0], "value": parts[1], "keep": keep})

      keep = calc_keep(ifblocks)
    elif str.startswith("#else"):
      ifblocks[-1]["keep"] = not ifblocks[-1]["keep"]

      keep = calc_keep(ifblocks)
    elif str.startswith("#endif"):
      if len(ifblocks) == 0:
        raise EspHomePreprocessorTemplateError(device["template"], linenumber, "endif nesting error, template tried closing block when no blocks are open.")

      ifblocks.pop()

      keep = calc_keep(ifblocks)
    elif str.startswith("#info"):
      if keep:
        parts = str.split()
        info(" ".join(parts[1:]))
    elif str.startswith("#warn"):
      if keep:
        parts = str.split()
        warn(" ".join(parts[1:]))
    elif str.startswith("#error"):
      if keep:
        parts = str.split()
        raise EspHomePreprocessorTemplateError(device["template"], linenumber, " ".join(parts[1:]))
    else:
      if keep:
        vars = re.findall(r"{#([^}]+)}", line)

        for var in vars:
          full_var = var
          var_parts = var.split()
          var_default = None
          var_error = None
          var_drop = False

          if len(var_parts) > 1:
            var = var_parts[0]

            if var_parts[1].lower() == "or":
              if len(var_parts) > 2:

                if var_parts[2].lower() == "default":
                  if len(var_parts) < 4:
                    raise EspHomePreprocessorTemplateError(device["template"], linenumber, "Template has tag '{}' with 'or' modifier and 'default' instruction but no default value".format(var))

                  var_default = " ".join(var_parts[3:])
                elif var_parts[2].lower() == "error":
                  if len(var_parts) < 4:
                    raise EspHomePreprocessorTemplateError(device["template"], linenumber, "Template has tag '{}' with 'or' modifier and 'error' instruction but no error message".format(var))

                  var_error = " ".join(var_parts[3:])
                elif var_parts[2].lower() == "drop":
                  var_drop = True
                else:
                  raise EspHomePreprocessorTemplateError(device["template"], linenumber, "Template has tag '{}' with 'or' modifier but unknown instruction '{}'".format(var, var_parts[2]))

              else:
                raise EspHomePreprocessorTemplateError(device["template"], linenumber, "Template has tag '{}' with 'or' modifier but no instruction".format(var))
            else:
              raise EspHomePreprocessorTemplateError(device["template"], linenumber, "Template has tag '{}' with unnwon modifier '{}'".format(var, var_parts[1]))

          try:
            value = device["defines"][var]
          except KeyError:
            if var_default is not None:
              value = var_default
            elif var_error is not None:
              raise EspHomePreprocessorTemplateError(device["template"], linenumber, var_error)
            elif var_drop:
              value = None
            else:
              raise EspHomePreprocessorTemplateError(device["template"], linenumber, "Template expects a define named '{}' but device didn't define it".format(var))

          tag = "{#"+full_var+"}"
          pos = line.index(tag)

          if value is None:
            line = ""
          else:
            if type(value) is str:
              pass
            elif type(value) in [list, tuple]:
              value = ordered_dump(value, default_flow_style=False)

            #if tag was alone, we must indent it properly
            if str == tag:
              line = ""
              for value_line in value.splitlines():
                line += " "*pos + value_line + "\n"
            else:
              line = line.replace(tag, value)

        output.append(line.rstrip())

  if len(ifblocks) != 0:
    raise EspHomePreprocessorTemplateError(device["template"], linenumber, "ifdef or ifndef nesting error, template processing finished with unclosed blocks.")

  return cleanup_template(output)

def cleanup_template(output):
  #Remove all consecutive newlines
  deletes = []
  prevline = ""
  for linenumber, line in enumerate(output):
    line = line.strip()
    if (prevline == "" and line == ""):
      deletes.append(linenumber)

    prevline = line

  for delete in reversed(deletes):
    del output[delete]

  return output

def calc_keep(ifblocks):
  for ifblock in ifblocks:
    if not ifblock["keep"]:
      return False

  return True

def save_output(output, device):
  directory = device["output_directory"]

  if not os.path.isdir(directory):
   raise EspHomePreprocessorError("Output directory '{0}' does not exist.".format(directory))

  try:
    filename = directory + os.path.sep + device["output_filename"]
  except KeyError as ex:
    filename = directory + os.path.sep + device["defines"]["id"] + ".yaml"

  log("Writing output to file '{0}'...".format(filename))

  with open(filename, 'w') as f:
    for line in output:
      f.write("%s\n" % line)

  log("Done!")

if __name__ == "__main__":
  sys.exit(main())